`timescale 1ns/1ps

`define wbclk_period 100
`define async_clk_period 63

`include "output/vlog_constants.v"

module main;
   reg clk=1;
   reg ram1_clk = 1;
   reg [31:0] rval;
   
   
   reg rst=0;

// generate clocks & reset
   always #(`wbclk_period/2) clk <= ~clk;
   always #(`async_clk_period/2) ram1_clk <= ~ram1_clk;

   initial #1000 rst <= 1;

`include "wishbone_stuff.v"


   
   reg [7:0]   ram1_addr = 0;
   reg [31:0]  ram1_data_wr = 0;
   wire [31:0] ram1_data_rd;
   reg [3:0]   ram1_bwsel = 4'b1111;
   reg 	       ram1_wr = 0, ram1_rd = 0;
   reg [9:0]   ram2_addr = 0;
   wire [15:0] ram2_data_rd;
   reg 	       ram2_rd = 0;

    task ram1_read;
      input[31:0] addr;
      output [31:0] data;
      begin
	 
 	 
	 ram1_addr = {2'b00, addr[9:2]};
	 ram1_rd = 1;

	 @(posedge ram1_clk);

	 
	 #1 data  = ram1_data_rd;
	 ram1_rd = 0;
	 
	 
      end
   endtask // wb_write

  task ram1_write;
      input[31:0] addr;
      input [31:0] data;
      begin
	 
	 #1 ram1_addr = {2'b00, addr[9:2]};
	 ram1_data_wr = data;
	 ram1_wr = 1;
 
 	 @(posedge ram1_clk);

	 #1 ram1_wr = 0;

      end
   endtask


   task ram2_read;
      input[31:0] addr;
      output [31:0] data;
      begin
	 
 	 
	 #1 ram2_addr = {2'b00, addr[11:2]};
	 ram2_rd = 1;

	 @(posedge clk);

	 
	 #1 data  = ram2_data_rd;
	 ram2_rd = 0;
	 
	 
      end
   endtask // wb_write
   
  wb_slave_test_rams 
    dut(
	.rst_n_i   (rst),
	.wb_clk_i  (clk),
	.wb_addr_i (wb_addr[10:0]),
	.wb_data_i (wb_data_o),
	.wb_data_o (wb_data_i),
	.wb_cyc_i  (wb_cyc),
	.wb_sel_i  (wb_bwsel),
	.wb_stb_i  (wb_stb),
	.wb_we_i   (wb_we),
	.wb_ack_o  (wb_ack),
	.clk1_i    (ram1_clk),

	.rams_mem1k_addr_i  (ram1_addr),
	.rams_mem1k_data_o  (ram1_data_rd),
	.rams_mem1k_rd_i    (ram1_rd),
	.rams_mem1k_data_i  (ram1_data_wr),
	.rams_mem1k_wr_i    (ram1_wr),
	.rams_mem1k_bwsel_i (ram1_bwsel),         
	.rams_mem2k_addr_i  (ram2_addr),
	.rams_mem2k_data_o  (ram2_data_rd),      
	.rams_mem2k_rd_i    (ram2_rd)      
  );
   
   integer     i;
   integer fail = 0;

      
   initial begin
      #2000; @(posedge clk); #1; // wait until the DUT is reset       

      $display("Test simple bus reads/writes...");

      wb_verbose(1);  
      
      wb_write(`BASE_RAMS_MEM1K, 32'hdeadbeef);
      wb_write(`BASE_RAMS_MEM1K + 4, 32'hcafecafe);
      wb_write(`BASE_RAMS_MEM1K + 'h200, 32'hfacedead);


      wb_read(`BASE_RAMS_MEM1K, rval);  if(rval != 'hdeadbeef) fail = 1;    
      wb_read(`BASE_RAMS_MEM1K + 4, rval);if(rval != 'hcafecafe) fail = 1;
      wb_read(`BASE_RAMS_MEM1K + 'h200, rval);if(rval != 'hfacedead) fail = 1;

      
      $display("Test mirrored bus reads/writes...");

      wb_write(`BASE_RAMS_MEM1K + 16, 32'h55555555);
      wb_read(`BASE_RAMS_MEM1K + 16, rval);if(rval != 'h55555555) fail = 1;
      
      wb_write(`BASE_RAMS_MEM1K + 4*`SIZE_RAMS_MEM1K + 16, 32'haaaaaaaa);
      wb_read(`BASE_RAMS_MEM1K + 16, rval);if(rval != 'haaaaaaaa) fail = 1;


      $display("Byte-access test...");

      wb_verbose(0);
      
      
      for(i=0;i<32;i=i+1) wb_write_byte(`BASE_RAMS_MEM1K + i, i + 1);
      for(i=0;i<32;i=i+1) begin
	wb_read_byte(`BASE_RAMS_MEM1K + i, rval);
	 if(rval != i+1)
	   fail = 1;
      end

   
      
      
      $display("mem1k: Bus write/mem read test...");
      
      for(i=0;i<256;i=i+1) wb_write(`BASE_RAMS_MEM1K + i*4, i+1);

      @(posedge ram1_clk); // sync ourselves to the edge of RAM1 async clock

      for(i=0;i<256;i=i+1) begin 
	 ram1_read(i*4, rval);
	 if(rval != i+1) fail =1;
      end

      $display("mem1k: Mem write/bus read test...");
     
      for(i=0;i<256;i=i+1) 
	ram1_write(i*4, 257-i);

      @(posedge clk); // sync back to wb clock
      

      for(i=0;i<256;i=i+1) begin
	 wb_read(`BASE_RAMS_MEM1K+i*4, rval);

	 if(rval != 257-i)
	    fail =1;

      end
      

      $display("mem2k: Bus write/mem read test...");

      
      for(i=0;i<512;i=i+1) wb_write(`BASE_RAMS_MEM2K + i*4, 113*i+41);

      @(posedge clk); // sync back to wb clock

      
      for(i=0;i<512;i=i+1) begin 
	 ram2_read(i*4, rval);

	 
	 if(rval != 113*i+41) 
	    fail =1;
      end

      
      if(fail)
	$display("TESTS FAILED");
      else
	$display("TESTS PASSED");
   end


endmodule
