#!/usr/bin/lua

-- wbgen2, (c) 2010 Tomasz Wlostowski
--                  CERN BE-Co-HT
-- LICENSED UNDER GPL v2

fieldtype_2_vhdl={};
fieldtype_2_vhdl[BIT]="std_logic";
fieldtype_2_vhdl[MONOSTABLE]="std_logic";
fieldtype_2_vhdl[SIGNED] = "signed";
fieldtype_2_vhdl[UNSIGNED] = "unsigned";
fieldtype_2_vhdl[ENUM] = "std_logic_vector";
fieldtype_2_vhdl[SLV] = "std_logic_vector";

cvtfunc_hdl = {};
cvtfunc_hdl[SLV] = "";
cvtfunc_hdl[SIGNED] = "signed";
cvtfunc_hdl[UNSIGNED] = "unsigned";


-------------------------------
-- GENERAL-PURPOSE FUNCTIONS --
-------------------------------

-- variable containing the current snippet of VHDL code
code_vhdl = {};

-- function emits a line of VHDL code.
function vhdl(s)
    table.insert(code_vhdl, s);
end

-- function forces single-tab left indentation
function vhdl_indent_left()
    table.insert(code_vhdl,"--@il");
end

-- function forces single-tab left indentation
function vhdl_indent_right()
    table.insert(code_vhdl,"--@ir");
end

-- function creates an empty snippet of VHDL code. The previous snippet is destroyed.
function vhdl_new_code()
    code_vhdl = {};
end

-- function pastes code into current snippet from another snippet.
function vhdl_paste_code(code)
    local i,v;

    if(code == nil) then  return; end
    
    for i,v in pairs(code) do
	table.insert(code_vhdl, v);
    end
end

-- generates new VHDL component instance
function vhdl_new_instance()
  local inst = {};
  
  inst.component = "";
  inst.name = "";
  inst.pmaps= {};
  inst.gmaps = {};

  return inst;
end

-------------------------------
-- VHDL CODE GENERATOR       --
-------------------------------


function gen_vhdl_signals_ram(reg)
    local prefix = string.lower(periph.hdl_prefix.."_"..reg.hdl_prefix);
    local port = {};
    
    if(reg.__type ~= TYPE_RAM) then
	return;
    end

    table.insert(hdl_ports, { name = prefix.."_addr_i"; length = reg.addr_bits - reg.wrap_bits; type = SLV; dir = "in"; });

    if(match(reg.access_dev, {READ_ONLY, READ_WRITE})) then
        table.insert(hdl_ports, { name = prefix.."_data_o"; length = reg.width; type = SLV; dir = "out"; });
    end    

    if(match(reg.access_dev, {WRITE_ONLY, READ_WRITE})) then
        table.insert(hdl_ports, { name = prefix.."_data_i"; length = reg.width; type = SLV; dir = "in"; });
        table.insert(hdl_ports, { name = prefix.."_wr_i"; length = 1; type = BIT; dir = "in"; });
    end
end


function gen_vhdl_signals_fifo(field, reg)
    local prefix = string.lower(periph.hdl_prefix.."_"..reg.hdl_prefix.."_"..field.hdl_prefix);

    if(reg.__type == TYPE_FIFO) then
	local suffix = csel(reg.direction == FROM_WB, "_o" , "_i");
	local mdir = csel(reg.direction == FROM_WB, "out" , "in");
        table.insert(hdl_ports, { name = prefix..suffix; length = field.size; type = field.type; dir = mdir; });
    end    
end


function gen_vhdl_controls_fifo(reg)
    local prefix = string.lower(periph.hdl_prefix.."_"..reg.hdl_prefix);

    if(reg.__type == TYPE_FIFO) then
	if flag_set(reg.fifo_flags_dev, FIFO_FULL) then		table.insert(hdl_ports, { name = prefix.."_ctl_full_o"; length = 1; type = BIT; dir = "out"; } ); end 
	if flag_set(reg.fifo_flags_dev, FIFO_EMPTY) then	table.insert(hdl_ports, { name = prefix.."_ctl_empty_o"; length = 1; type = BIT; dir = "out"; } ); end 
	if flag_set(reg.fifo_flags_dev, FIFO_ALMOST_FULL) then  table.insert(hdl_ports, { name = prefix.."_ctl_almostfull_o"; length = 1; type = BIT; dir = "out"; } ); end 
	if flag_set(reg.fifo_flags_dev, FIFO_ALMOST_EMPTY) then table.insert(hdl_ports, { name = prefix.."_ctl_almostempty_o"; length = 1; type = BIT; dir = "out"; } ); end 
	if flag_set(reg.fifo_flags_dev, FIFO_CLEAR) then 	table.insert(hdl_ports, { name = prefix.."_ctl_clear_i"; length = 1; type = BIT; dir = "in"; } ); end 
    end    
end

-- generates nice VHDL file header
function gen_vhdl_header(file)
    vhdl_new_code();
    vhdl ("---------------------------------------------------------------------------------------");
    vhdl ("-- Title          : Wishbone slave core for "..periph.name);
    vhdl ("---------------------------------------------------------------------------------------");
    vhdl ("-- File           : "..output_vhdl_file);
    vhdl ("-- Author         : auto-generated by wbgen2 from "..input_wb_file);
    vhdl ("-- Created        : "..os.date());
    vhdl ("-- Standard       : VHDL'87");
    vhdl ("---------------------------------------------------------------------------------------");
    vhdl ("-- THIS FILE WAS GENERATED BY wbgen2 FROM SOURCE FILE "..input_wb_file);
    vhdl ("-- DO NOT HAND-EDIT UNLESS IT'S ABSOLUTELY NECESSARY!");
    vhdl ("---------------------------------------------------------------------------------------");
    vhdl ("");
    vhdl ("");
    vhdl ("library ieee;");
    vhdl ("use ieee.std_logic_1164.all;");
    vhdl ("use ieee.numeric_std.all;");
    vhdl ("");
--    vhdl ("library wbgen2;");
--    vhdl ("use wbgen2.wbgen2_components.all;");
    vhdl ("");
    vhdl ("library work;");
    vhdl ("");
    return code_vhdl;
end

-- generates a string containing VHDL-compatible binary numeric value of size numbits
function gen_vhdl_bin_literal(value, numbits)
    local str ='\"';
    local i,n,d,r;
    
    n=value;
    r=math.pow(2, numbits-1);
    
    for i=1,numbits do
	d=math.floor(n/r);
	str=str..csel(d>0,"1","0");
	n=n%r;
	r=r/2;
    end
    return str..'\"';
end

function gen_vhdl_block_address_decoder()
    local s = {};
    s= "\n\n-- BLOCK ADDRESS DECODER --\n";
    s=s.."address_decode_1: process (wb_addr_i) begin\n";
    s=s.."\treg_blk_sel <= '1' when (wb_addr_i("..(address_bus_width-1).." downto "..(address_bus_width - address_bus_select_bits)..") = "..gen_vhdl_bin_literal(0, address_bus_select_bits)..") else '0';\n";

    foreach_reg(function(reg) 
	if(reg.__type == TYPE_RAM) then
	    s=s.."\t"..reg.hdl_prefix.."_blk_sel <= '1' when (wb_addr_i("..(address_bus_width-1).." downto "..(address_bus_width - address_bus_select_bits)..") = "..gen_vhdl_bin_literal(reg.select_bits, address_bus_select_bits)..") else '0';\n";
	end
    end );
    s=s.."end process;\n";
    return s;
end


-- constructor for a VHDL signal
function signal(type, nbits, name, comment)
    local t = {}
    t.comment = comment;
    t.type = type;
    t.range= nbits;
    t.name = name;
    return t;
end

-- constructor for a VHDL port
function port(type, nbits, dir, name, comment)
    local t = {}
	t.comment = comment;
    t.type = type;
    t.range= nbits;
    t.name = name;
    t.dir = dir;
    return t;
end

function vhdl_portmap(instance, port, sig, range_port, range_sig)
  local map = {};
  map.type = "portmap";
  map.port = port;
  map.signal = sig;
  map.range_port = range_port;
  map.range_sig = range_sig;
  table_join(instance.pmaps, map);
end

function vhdl_generic(instance, gname, gvalue)
  local gen = {};

  gen.type = "genmap";
  gen.name = gname;
  gen.value = gvalue;
  table_join(instance.gmaps, gen);
end


function gen_vhdl_type(field)
	local s = fieldtype_2_vhdl[field.type];
	
	if(field.type == SLV or field.type == UNSIGNED or field.type == SIGNED) then
		s=s.."("..(field.size-1).." downto 0)";
	end
	return s;
end

function write_vhdl_code(code, filename)
local cur_indent = 0;
local file = io.open(filename, "w");
local i,v,j;
--local in_case;

for i,v in pairs(code) do
    

    if(v == "--@il") then
	cur_indent = cur_indent - 1;
    elseif ( v == "--@ir") then
    	cur_indent = cur_indent + 1;
    else

    if(
	string.find(v, "^else") ~= nil or 
	string.find(v, "^elsif") ~= nil or 
	string.find(v, "^end if;") ~= nil or 
---	string.find(v, "^else") ~= nil or 
	string.find(v, "^end case;") ~= nil or 
	string.find(v, "^end process;") ~= nil or 
	string.find(v, "^end;") ~= nil or 
	string.find(v, "^when") ~= nil or 
	string.find(v, "^end ") ~= nil) then
	cur_indent = cur_indent -1;
    end

    if(string.find(v,"^%-%-") == nil) then	
	for j=1,cur_indent do file.write(file, "    "); end
    end
--    print (i,v);    
    
    file.write(file, v.."\n");



    if(
	string.find(v, "^begin") ~= nil or 
	string.find(v, "^elsif") ~= nil or 
	string.find(v, "^else") ~= nil or 
--	string.find(v, "^process") ~= nil or 
	string.find(v, "^case") ~= nil or 
	string.find(v, "^when") ~= nil or 
	string.find(v, "^if") ~= nil or 
	string.find(v, "^entity") ~= nil or
	string.find(v, "^architecture") ~= nil or
	string.find(v, "^port") ~= nil or
	string.find(v, "^generic") ~= nil 
	
	) then
	cur_indent = cur_indent + 1;
	end

    end
    end
io.close(file);
end

function gen_vhdl_wishbone_ports()

    local ports = {
	port(BIT, 0, "in", "rst_n_i"),
	port(BIT, 0, "in", "wb_clk_i"),
	};
	
    if(address_bus_width > 0 ) then
	table_join(ports, { port(SLV, address_bus_width, "in", "wb_addr_i") });
    end

    table_join(ports, {
	port(SLV, DATA_BUS_WIDTH, "in", "wb_data_i"),
	port(SLV, DATA_BUS_WIDTH, "out", "wb_data_o"),
	
	port(BIT, 0, "in", "wb_cyc_i"),
	port(BIT, 0, "in", "wb_sel_i"),
	port(BIT, 0, "in", "wb_stb_i"),
	port(BIT, 0, "in", "wb_we_i"),
	port(BIT, 0, "out", "wb_ack_o")
    });
    
    return ports;
end


function vhdl_build_clock_list()
    local allclocks = tree_2_table("clock");
    local i,v;
    local clockports = {};
    
    remove_duplicates(allclocks);
    
    for i,v in pairs(allclocks) do
    	table.insert(clockports, port(BIT, 0, "in", v));
    end

    return clockports;
end

function gen_vhdl_entity()
    local portlist = {};
    local i, last;

    vhdl_new_code();
    vhdl ("entity "..periph.hdl_entity.." is");
    vhdl ("port (");

-- generate a global port table by joining port tables for all fields/registers
    table_join(portlist, gen_vhdl_wishbone_ports());
    table_join(portlist, vhdl_build_clock_list());
    table_join(portlist, tree_2_table("ports"));

    vhdl_build_clock_list();

-- output the lines into file
    for i=1,table.getn(portlist) do
	local port = portlist[i];
    
        if(i == table.getn(portlist)) then
	    last = true;
	else
	    last = false;
        end

	if(port.comment ~= nil) then
	  vhdl("-- "..port.comment);
	end
	

	local line = string.format("%-40s : %-6s %s", port.name, port.dir, fieldtype_2_vhdl[port.type]);


        if(port.range > 0) then
    	    line = line.."("..(port.range-1).." downto 0)";
        end    
        
	line=line..csel(last, "", ";");
	vhdl(line);
    end

    
    vhdl_indent_left();
    vhdl(");");
    vhdl_indent_left();
    vhdl("end "..periph.hdl_entity..";");
    vhdl("");
    vhdl("");
    vhdl("architecture behavioral of "..periph.hdl_entity.." is ");
    
    return code_vhdl;

end

function gen_vhdl_signals()
	local siglist = {};
	local i,v;
	local s;
	
	siglist = tree_2_table("signals");
	
	wb_sigs = { 	    signal(BIT, 0, "wb_ack_regbank"),
			    signal(UNSIGNED, 4, "ack_cntr"),
			    signal(BIT,0, "ack_in_progress"),
			    signal(BIT,0, "tmpbit"),
			    signal(SLV, DATA_BUS_WIDTH, "wb_data_out_int") };
	
	
	table_join(siglist, wb_sigs);
	
	vhdl_new_code();
	
	for i,v in pairs (siglist) do
--    		print (fieldtype_2_vhdl[v.type].." "..v.name);
--    		print (v.type);
		s=string.format("signal %-40s : %-15s", v.name, fieldtype_2_vhdl[v.type]);
		if(v.range > 0) then
			s=s..string.format("(%d downto 0)", v.range-1);
		end
		s=s..";";
		vhdl(s);
	end
	
	
	return code_vhdl;
end

function vhdl_generate_constants()
	vhdl_new_code();

  foreach_reg(function(reg) 
	if(reg.__type == TYPE_REG) then
		vhdl(string.format("constant %-30s : std_logic_vector(%d downto 0) := %-10s;", "ADDR_"..string.upper(periph.hdl_prefix.."_"..reg.hdl_prefix), address_bus_width-1, gen_vhdl_bin_literal(reg.base, address_bus_width)));
	end
	end
	);
	
	vhdl("");
	return code_vhdl;
end

function gen_vhdl_code(filename)
	local code_header = gen_vhdl_header();
	local code_regbank = gen_vhdl_regbank();    
	local code_rams = gen_vhdl_rams();
	local code_signals = gen_vhdl_signals();
	local code_entity = gen_vhdl_entity();
  

	local code = {};

	table_join(code, code_header);
	table_join(code, code_entity);

	if(vhdl_gen_reg_constants == true) then
		table_join(code, vhdl_generate_constants());
	end

	table_join(code, code_signals);
	table_join(code, code_regbank);
	table_join(code, code_rams);

	vhdl_new_code();
	vhdl ("wb_data_o <= wb_data_out_int;");
	vhdl ("wb_ack_o <= wb_ack_regbank;");
	vhdl ("end behavioral;");
	
	table_join(code, code_vhdl);

	write_vhdl_code(code, filename);
end

function gen_vlog_constants(filename)
	print("genvlog: "..filename);
	local file = io.open(filename, "w");
 
 	if(file == nil) then
 		die("can't open "..filename.." for writing.");
 	end
	 
	  foreach_reg(function(reg) 
		if(reg.__type == TYPE_REG) then
			file.write(file, string.format("`define %-30s %d'h%x\n", "ADDR_"..string.upper(periph.hdl_prefix.."_"..reg.hdl_prefix), address_bus_width, reg.base));
		end
		end
		);

	io.close(file);
end